(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: DepthOfField.cpp
//
// Desc: Example code showing how to do a depth of field effect.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
unit DepthOfFieldUnit;

{$I DirectX.inc}

interface

uses
  Windows, Messages, Direct3D9, D3DX9,
  D3DApp, D3DFont, D3DUtil, DXUtil,
  Resource;

{.$DEFINE DEBUG_VS}   // Uncomment this line to debug vertex shaders
{.$DEFINE DEBUG_PS}   // Uncomment this line to debug pixel shaders

type
  TVertex = packed record
    pos: TD3DXVector4;
    clr: DWORD;
    tex1: TD3DXVector2;
    tex2: TD3DXVector2;
    tex3: TD3DXVector2;
    tex4: TD3DXVector2;
    tex5: TD3DXVector2;
    tex6: TD3DXVector2;
  end;

const
  TVertex_FVF = D3DFVF_XYZRHW or D3DFVF_DIFFUSE or D3DFVF_TEX6;

type
  CMyD3DApplication = class(CD3DApplication)
  private
    m_Camera                   : CFirstPersonCamera;
    m_pFont                    : ID3DXFont;           // Font for drawing text
    m_fAspectRatio             : Single;

    m_Vertex: array[0..3] of TVertex;

    m_pFullScreenTexture       : IDirect3DTexture9;
    m_pRenderToSurface         : ID3DXRenderToSurface;
    m_pFullScreenTextureSurf   : IDirect3DSurface9;

    m_pScene1Mesh              : ID3DXMesh;
    m_pScene1MeshTexture       : IDirect3DTexture9;
    m_pScene2Mesh              : ID3DXMesh;
    m_pScene2MeshTexture       : IDirect3DTexture9;
    m_nCurrentScene            : Integer;
    m_pEffect                  : ID3DXEffect;

    m_vFocalPlaneDest          : TD3DXVector4;
    m_vFocalPlaneSrc           : TD3DXVector4;
    m_vFocalPlaneCur           : TD3DXVector4;
    m_fChangeTime              : Single;
    m_bShowBlurFactor          : BOOL;
    m_bShowUnblurred           : BOOL;
    m_bShowHelp                : BOOL;
    m_dwBackgroundColor        : DWORD;

    m_ViewportFB               : TD3DViewport9;
    m_ViewportOffscreen        : TD3DViewport9;

    m_fBlurConst               : Single;

    m_TechniqueIndex           : DWORD;

    m_hFocalPlane              : TD3DXHandle;
    m_hWorld                   : TD3DXHandle;
    m_hWorldView               : TD3DXHandle;
    m_hWorldViewProjection     : TD3DXHandle;
    m_hMeshTexture             : TD3DXHandle;
    m_hTechWorldWithBlurFactor : TD3DXHandle;
    m_hTechShowBlurFactor      : TD3DXHandle;
    m_hTechShowUnmodified      : TD3DXHandle;
    m_hTech: array[0..4] of TD3DXHandle;

  protected
    function InitDeviceObjects: HRESULT; override;
    function RestoreDeviceObjects: HRESULT; override;
    function InvalidateDeviceObjects: HRESULT; override;
    function DeleteDeviceObjects: HRESULT; override;
    function FinalCleanup: HRESULT; override;
    function Render: HRESULT; override;
    function FrameMove: HRESULT; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;

    procedure RenderText;
    function LoadMesh(strFileName: PChar; out ppMesh: ID3DXMesh): HRESULT;
    procedure SetupQuad;
    function  UpdateTechniqueSpecificVariables: HRESULT;

  public
    constructor Create;
    
    function MsgProc(hWnd: HWND; uMsg: LongWord; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  end;

implementation

uses Math, SysUtils;

//-----------------------------------------------------------------------------
// Technique list
//-----------------------------------------------------------------------------
const
  m_TechniqueCount = 5;
  m_TechniqueNames: array[0..m_TechniqueCount - 1] of PAnsiChar = (
    'UsePS20ThirteenLookups',
    'UsePS20SevenLookups',
    'UsePS20SixTexcoords',
    'UsePS11FourTexcoordsNoRings',
    'UsePS11FourTexcoordsWithRings'
  );

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;
  m_Camera:= CFirstPersonCamera.Create;

  m_strWindowTitle    := 'Depth of Field';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_pFont := nil;
  m_fAspectRatio := 0.0;
  m_dwCreationWidth  := 640;
  m_dwCreationHeight := 480;

  m_pFullScreenTexture := nil;
  m_pFullScreenTextureSurf := nil;
  m_pRenderToSurface := nil;
  m_pEffect := nil;

  m_vFocalPlaneDest := D3DXVector4(0.0, 0.0, 1.0, -2.5);
  m_vFocalPlaneSrc := m_vFocalPlaneDest;
  m_fChangeTime := 0.0;

  m_pScene1Mesh := nil;
  m_pScene1MeshTexture := nil;
  m_pScene2Mesh := nil;
  m_pScene2MeshTexture := nil;
  m_nCurrentScene := 1;

  m_bShowBlurFactor := False;
  m_bShowUnblurred := False;
  m_bShowHelp := True;
  m_dwBackgroundColor := $00003F3F;

  m_fBlurConst := 4.0;
  m_TechniqueIndex := 0;

  m_hFocalPlane := nil;
  m_hWorld := nil;
  m_hWorldView := nil;
  m_hWorldViewProjection := nil;
  m_hMeshTexture := nil;
  m_hTechWorldWithBlurFactor := nil;
  m_hTechShowBlurFactor := nil;
  m_hTechShowUnmodified := nil;
  ZeroMemory(@m_hTech, SizeOf(m_hTech));
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: This creates all device-dependent managed objects, such as managed
//       textures and managed vertex buffers.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HRESULT;
var
  hDC: Windows.HDC;
  nHeight: Integer;
  str: array[0..MAX_PATH-1] of Char;
  hr: HRESULT;
  dwShaderFlags: DWORD;
  pBufferErrors: ID3DXBuffer;
begin
  // Initialize the font
  hDC := GetDC(0);
  nHeight := -MulDiv(9, GetDeviceCaps(hDC, LOGPIXELSY), 72);
  ReleaseDC(0, hDC);
  Result:= D3DXCreateFont(m_pd3dDevice, nHeight, 0, FW_BOLD, 0, FALSE, DEFAULT_CHARSET,
                          OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH or FF_DONTCARE,
                          'Arial', m_pFont);
  if FAILED(Result) then Exit;

  // Load the meshs
  Result := LoadMesh('tiger.x', m_pScene1Mesh);
  if FAILED(Result) then Exit;
  Result := LoadMesh('sphere.x', m_pScene2Mesh);
  if FAILED(Result) then Exit;

  // Load the mesh textures
  DXUtil_FindMediaFileCb(str, SizeOf(str), 'tiger.bmp');
  hr := D3DXCreateTextureFromFile(m_pd3dDevice, str, m_pScene1MeshTexture);
  if FAILED(hr) then m_pScene1MeshTexture := nil;

  DXUtil_FindMediaFileCb(str, sizeof(str), 'earth.bmp');
  hr := D3DXCreateTextureFromFile(m_pd3dDevice, str, m_pScene2MeshTexture);
  if FAILED(hr) then m_pScene2MeshTexture := nil;

  // Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders with the shader debugger.
  // Debugging vertex shaders requires either REF or software vertex processing, and debugging
  // pixel shaders requires REF.  The D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug
  // experience in the shader debugger.  It enables source level debugging, prevents instruction
  // reordering, prevents dead code elimination, and forces the compiler to compile against the next
  // higher available software target, which ensures that the unoptimized shaders do not exceed
  // the shader model limitations.  Setting these flags will cause slower rendering since the shaders
  // will be unoptimized and forced into software.  See the DirectX documentation for more information
  // about using the shader debugger.
  dwShaderFlags := 0;
  {$IFDEF DEBUG_VS}
    dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
    dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
  {$ENDIF}

  // Load the D3DX effect file into a ID3DXEffect interface.
  pBufferErrors := nil;
  DXUtil_FindMediaFileCb(str, SizeOf(str), 'DepthOfField.fx');
  Result := D3DXCreateEffectFromFile(m_pd3dDevice, str, nil, nil, dwShaderFlags, nil, m_pEffect, @pBufferErrors);
  // if FAILED(Result) then Exit;
end;


//-----------------------------------------------------------------------------
// Name: LoadMesh()
// Desc: Loads a mesh, ensures it has normals, and optimizes it for the
//       vertex cache size of the current card.
//-----------------------------------------------------------------------------
function CMyD3DApplication.LoadMesh(strFileName: PChar; out ppMesh: ID3DXMesh): HRESULT;
var
  pMesh: ID3DXMesh;
  str: array[0..MAX_PATH-1] of Char;
  rgdwAdjacency: PDWORD;
  pTempMesh: ID3DXMesh;
begin
  pMesh := nil;
  if (@ppMesh = nil) then
  begin
    Result:= E_INVALIDARG;
    Exit;
  end;

  DXUtil_FindMediaFileCb(str, SizeOf(str), strFileName);
  Result := D3DXLoadMeshFromX(str, D3DXMESH_MANAGED,
                              m_pd3dDevice, nil, nil, nil, nil, pMesh);
  if FAILED(Result) or (pMesh = nil) then Exit;

  //rgdwAdjacency := nil; //BAA: never used anyway

  // Make sure there are normals which are required for lighting
  if ((pMesh.GetFVF and D3DFVF_NORMAL) = 0) then
  begin
    Result := pMesh.CloneMeshFVF(pMesh.GetOptions,
                                 pMesh.GetFVF or D3DFVF_NORMAL,
                                 m_pd3dDevice, pTempMesh);
    if FAILED(Result) then Exit;

    D3DXComputeNormals(pTempMesh, nil);

    SAFE_RELEASE(pMesh);
    pMesh := pTempMesh;
  end;

  // Optimze the mesh to make it fast for the user's graphics card
  try
    GetMem(rgdwAdjacency, SizeOf(DWORD) * pMesh.GetNumFaces * 3);
    // if (rgdwAdjacency = nil) then Result:= E_OUTOFMEMORY;
  except
    Result:= E_OUTOFMEMORY;
    Exit;
  end;
  pMesh.ConvertPointRepsToAdjacency(nil, rgdwAdjacency);
  pMesh.OptimizeInplace(D3DXMESHOPT_VERTEXCACHE, rgdwAdjacency, nil, nil, nil);
  FreeMem(rgdwAdjacency);

  ppMesh := pMesh;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: SetupQuad()
// Desc:
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.SetupQuad;
var
  desc: TD3DSurfaceDesc;
  fWidth5, fHeight5: Single;
  fHalf, fOffOne, fOffTwo: Single;
  fTexWidth1, fTexHeight1: Single;
  fWidthMod, fHeightMod: Single;
begin
  m_pFullScreenTextureSurf.GetDesc(desc);

  fWidth5 := m_d3dsdBackBuffer.Width - 0.5;
  fHeight5 := m_d3dsdBackBuffer.Height - 0.5;

  fHalf := m_fBlurConst;
  fOffOne := fHalf * 0.5;
  fOffTwo := fOffOne * Sqrt(3.0);

  fTexWidth1 := m_d3dsdBackBuffer.Width / desc.Width;
  fTexHeight1 := m_d3dsdBackBuffer.Height / desc.Height;

  fWidthMod := 1.0 / desc.Width;
  fHeightMod := 1.0 / desc.Height;

  // Create vertex buffer.
  // m_Vertex[0].tex1 == full texture coverage
  // m_Vertex[0].tex2 == full texture coverage, but shifted y by -fHalf*fHeightMod
  // m_Vertex[0].tex3 == full texture coverage, but shifted x by -fOffTwo*fWidthMod & y by -fOffOne*fHeightMod
  // m_Vertex[0].tex4 == full texture coverage, but shifted x by +fOffTwo*fWidthMod & y by -fOffOne*fHeightMod
  // m_Vertex[0].tex5 == full texture coverage, but shifted x by -fOffTwo*fWidthMod & y by +fOffOne*fHeightMod
  // m_Vertex[0].tex6 == full texture coverage, but shifted x by +fOffTwo*fWidthMod & y by +fOffOne*fHeightMod
  m_Vertex[0].pos := D3DXVector4(fWidth5, -0.5, 0.0, 1.0);
  m_Vertex[0].clr := D3DCOLOR_COLORVALUE(0.5, 0.5, 0.5, 0.66666);
  m_Vertex[0].tex1 := D3DXVector2(fTexWidth1, 0.0);
  m_Vertex[0].tex2 := D3DXVector2(fTexWidth1, 0.0 - fHalf*fHeightMod);
  m_Vertex[0].tex3 := D3DXVector2(fTexWidth1 - fOffTwo*fWidthMod, 0.0 - fOffOne*fHeightMod);
  m_Vertex[0].tex4 := D3DXVector2(fTexWidth1 + fOffTwo*fWidthMod, 0.0 - fOffOne*fHeightMod);
  m_Vertex[0].tex5 := D3DXVector2(fTexWidth1 - fOffTwo*fWidthMod, 0.0 + fOffOne*fHeightMod);
  m_Vertex[0].tex6 := D3DXVector2(fTexWidth1 + fOffTwo*fWidthMod, 0.0 + fOffOne*fHeightMod);

  m_Vertex[1].pos := D3DXVector4(fWidth5, fHeight5, 0.0, 1.0);
  m_Vertex[1].clr := D3DCOLOR_COLORVALUE(0.5, 0.5, 0.5, 0.66666);
  m_Vertex[1].tex1 := D3DXVector2(fTexWidth1, fTexHeight1);
  m_Vertex[1].tex2 := D3DXVector2(fTexWidth1, fTexHeight1 - fHalf*fHeightMod);
  m_Vertex[1].tex3 := D3DXVector2(fTexWidth1 - fOffTwo*fWidthMod, fTexHeight1 - fOffOne*fHeightMod);
  m_Vertex[1].tex4 := D3DXVector2(fTexWidth1 + fOffTwo*fWidthMod, fTexHeight1 - fOffOne*fHeightMod);
  m_Vertex[1].tex5 := D3DXVector2(fTexWidth1 - fOffTwo*fWidthMod, fTexHeight1 + fOffOne*fHeightMod);
  m_Vertex[1].tex6 := D3DXVector2(fTexWidth1 + fOffTwo*fWidthMod, fTexHeight1 + fOffOne*fHeightMod);

  m_Vertex[2].pos := D3DXVector4(-0.5, -0.5, 0.0, 1.0);
  m_Vertex[2].clr := D3DCOLOR_COLORVALUE(0.5, 0.5, 0.5, 0.66666);
  m_Vertex[2].tex1 := D3DXVector2(0.0, 0.0);
  m_Vertex[2].tex2 := D3DXVector2(0.0, 0.0 - fHalf*fHeightMod);
  m_Vertex[2].tex3 := D3DXVector2(0.0 - fOffTwo*fWidthMod, 0.0 - fOffOne*fHeightMod);
  m_Vertex[2].tex4 := D3DXVector2(0.0 + fOffTwo*fWidthMod, 0.0 - fOffOne*fHeightMod);
  m_Vertex[2].tex5 := D3DXVector2(0.0 - fOffTwo*fWidthMod, 0.0 + fOffOne*fHeightMod);
  m_Vertex[2].tex6 := D3DXVector2(0.0 + fOffTwo*fWidthMod, 0.0 + fOffOne*fHeightMod);

  m_Vertex[3].pos := D3DXVector4(-0.5, fHeight5, 0.0, 1.0);
  m_Vertex[3].clr := D3DCOLOR_COLORVALUE(0.5, 0.5, 0.5, 0.66666);
  m_Vertex[3].tex1 := D3DXVector2(0.0, fTexHeight1);
  m_Vertex[3].tex2 := D3DXVector2(0.0, fTexHeight1 - fHalf*fHeightMod);
  m_Vertex[3].tex3 := D3DXVector2(0.0 - fOffTwo*fWidthMod, fTexHeight1 - fOffOne*fHeightMod);
  m_Vertex[3].tex4 := D3DXVector2(0.0 + fOffTwo*fWidthMod, fTexHeight1 - fOffOne*fHeightMod);
  m_Vertex[3].tex5 := D3DXVector2(0.0 + fOffTwo*fWidthMod, fTexHeight1 + fOffOne*fHeightMod);
  m_Vertex[3].tex6 := D3DXVector2(0.0 - fOffTwo*fWidthMod, fTexHeight1 + fOffOne*fHeightMod);
end;

//-----------------------------------------------------------------------------
// Name: UpdateTechniqueSpecificVariables()
// Desc: Certain parameters need to be specified for specific techniques
//-----------------------------------------------------------------------------
function  CMyD3DApplication.UpdateTechniqueSpecificVariables: HRESULT;
type
  PD3DXVector2Array = ^TD3DXVector2Array;
  TD3DXVector2Array = array[0..MaxInt div SizeOf(TD3DXVector2)-1] of TD3DXVector2;
var
  strInputArrayName, strOutputArrayName: PAnsiChar;
  nNumKernelEntries: Integer;
  hAnnotation: TD3DXHandle;
  hTech: TD3DXHandle;
  aKernel: PD3DXVector2Array;
  
  desc: TD3DSurfaceDesc;
  fWidthMod, fHeightMod: Single;
  iEntry: Integer;
begin
  // Create the post-process quad and set the texcoords based on the blur factor
  SetupQuad;

  Result := S_FALSE;

  // Get the handle to the current technique
  hTech := m_pEffect.GetTechniqueByName(m_TechniqueNames[m_TechniqueIndex]);
  if(hTech = nil) then Exit; // This will happen if the technique doesn't have this annotation

  // Get the value of the annotation int named "NumKernelEntries" inside the technique
  hAnnotation := m_pEffect.GetAnnotationByName(hTech, 'NumKernelEntries');
  if (hAnnotation = nil) then Exit; // This will happen if the technique doesn't have this annotation
  Result := m_pEffect.GetInt(hAnnotation, nNumKernelEntries);
  if FAILED(Result) or (nNumKernelEntries <= 0) then Exit; // This shouldn't happen unless the effect file has been modified

  // Get the value of the annotation string named "KernelInputArray" inside the technique
  Result := S_FALSE;
  hAnnotation := m_pEffect.GetAnnotationByName(hTech, 'KernelInputArray');
  if (hAnnotation = nil) then Exit; // This will happen if the technique doesn't have this annotation
  Result := m_pEffect.GetString(hAnnotation, strInputArrayName);
  if FAILED(Result) then Exit;

  // Get the value of the annotation string named "KernelOutputArray" inside the technique
  Result := S_FALSE;
  hAnnotation := m_pEffect.GetAnnotationByName(hTech, 'KernelOutputArray');
  if (hAnnotation = nil) then Exit; // This will happen if the technique doesn't have this annotation
  Result := m_pEffect.GetString(hAnnotation, strOutputArrayName);
  if FAILED(Result) then Exit;

  // Create a array to store the input array
  try
    GetMem(aKernel, SizeOf(TD3DXVector2)*nNumKernelEntries);
    // if (aKernel = nil) Result:= E_OUTOFMEMORY;
  except
    Result:= E_OUTOFMEMORY;
    Exit;
  end;

  // Get the input array
  m_pEffect.GetValue(TD3DXHandle(strInputArrayName), aKernel, SizeOf(TD3DXVector2) * nNumKernelEntries);

  // Get the size of the texture
  m_pFullScreenTextureSurf.GetDesc(desc);

  // Calculate the scale factor to convert the input array to screen space
  fWidthMod := m_fBlurConst / desc.Width ;
  fHeightMod := m_fBlurConst / desc.Height;

  // Scale the effect's kernel from pixel space to tex coord space
  // In pixel space 1 unit = one pixel and in tex coord 1 unit = width/height of texture
  for iEntry := 0 to nNumKernelEntries - 1 do
  begin
    aKernel[iEntry].x := aKernel[iEntry].x * fWidthMod;
    aKernel[iEntry].y := aKernel[iEntry].y * fHeightMod;
  end;

  // Pass the updated array values to the effect file
  m_pEffect.SetValue(TD3DXHandle(strOutputArrayName), aKernel, SizeOf(TD3DXVector2)*nNumKernelEntries);

  FreeMem(aKernel); // SAFE_DELETE_ARRAY(aKernel);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HRESULT;
var
  vecEye, vecAt: TD3DXVector3;
  desc: TD3DSurfaceDesc;
  colorWhite, colorBlack, colorAmbient: TD3DXColor;
  i: Integer;
  OriginalTechnique: DWORD;
  hTech: TD3DXHandle;
begin
  if Assigned(m_pFont) then m_pFont.OnResetDevice;
  if Assigned(m_pEffect) then m_pEffect.OnResetDevice;

  // Setup the camera with view & projection matrix
  vecEye := D3DXVector3(1.3, 1.1, -3.3);
  vecAt := D3DXVector3(0.75, 0.9, -2.5);
  m_Camera.SetViewParams(vecEye, vecAt);
  m_fAspectRatio := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  m_Camera.SetProjParams(D3DXToRadian(60.0), m_fAspectRatio, 0.5, 100.0);
  m_Camera.SetResetCursorAfterMove(not m_bWindowed);

  m_pd3dDevice.GetViewport(m_ViewportFB);

  // Backbuffer viewport is identical to frontbuffer, except starting at 0, 0
  m_ViewportOffscreen := m_ViewportFB;
  m_ViewportOffscreen.X := 0;
  m_ViewportOffscreen.Y := 0;

  // Create fullscreen renders target texture
  Result := D3DXCreateTexture(m_pd3dDevice, m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, m_pFullScreenTexture);
  if FAILED(Result) then
  begin
    Result := D3DXCreateTexture(m_pd3dDevice, m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 1, 0, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, m_pFullScreenTexture);
    if FAILED(Result) then Exit;
  end;

  m_pFullScreenTexture.GetSurfaceLevel(0, m_pFullScreenTextureSurf);
  m_pFullScreenTextureSurf.GetDesc(desc);

  // Create a ID3DXRenderToSurface to help render to a texture on cards
  // that don't support render targets
  Result := D3DXCreateRenderToSurface(m_pd3dDevice, desc.Width, desc.Height,
              desc.Format, True, D3DFMT_D16, m_pRenderToSurface);
  if FAILED(Result) then Exit;

  // clear the surface alpha to 0 so that it does not bleed into a "blurry" background
  //   this is possible because of the avoidance of blurring in a non-blurred texel
  if SUCCEEDED(m_pRenderToSurface.BeginScene(m_pFullScreenTextureSurf, nil)) then
  begin
    m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET, $00, 1.0, 0);
    m_pRenderToSurface.EndScene(0);
  end;

  colorWhite := D3DXColor(1.0, 1.0, 1.0, 1.0);
  colorBlack := D3DXColor(0.0, 0.0, 0.0, 1.0);
  colorAmbient := D3DXColor(0.25, 0.25, 0.25, 1.0);

  // Get D3DXHANDLEs to the parameters/techniques that are set every frame so 
  // D3DX doesn't spend time doing string compares.  Doing this likely won't affect
  // the perf of this simple sample but it should be done in complex engine.
  m_hFocalPlane               := m_pEffect.GetParameterByName(nil, 'vFocalPlane');
  m_hWorld                    := m_pEffect.GetParameterByName(nil, 'mWorld');
  m_hWorldView                := m_pEffect.GetParameterByName(nil, 'mWorldView');
  m_hWorldViewProjection      := m_pEffect.GetParameterByName(nil, 'mWorldViewProjection');
  m_hMeshTexture              := m_pEffect.GetParameterByName(nil, 'MeshTexture');
  m_hTechWorldWithBlurFactor  := m_pEffect.GetTechniqueByName('WorldWithBlurFactor');
  m_hTechShowBlurFactor       := m_pEffect.GetTechniqueByName('ShowBlurFactor');
  m_hTechShowUnmodified       := m_pEffect.GetTechniqueByName('ShowUnmodified');
  for i:= 0 to 4 do
    m_hTech[i] := m_pEffect.GetTechniqueByName(m_TechniqueNames[i]);

  // Set the vars in the effect that doesn't change each frame
  m_pEffect.SetVector('MaterialAmbientColor', PD3DXVector4(@colorAmbient)^);
  m_pEffect.SetVector('MaterialDiffuseColor', PD3DXVector4(@colorWhite)^);
  m_pEffect.SetTexture('RenderTargetTexture', m_pFullScreenTexture);

  // Check if the current technique is valid for the new device/settings
  // Start from the current technique, increment until we find one we can use.
  OriginalTechnique := m_TechniqueIndex;
  repeat
    if (m_TechniqueIndex = m_TechniqueCount) then m_TechniqueIndex := 0;

    hTech := m_pEffect.GetTechniqueByName(m_TechniqueNames[m_TechniqueIndex]);
    if SUCCEEDED(m_pEffect.ValidateTechnique(hTech)) then Break;

    Inc(m_TechniqueIndex);
  until (OriginalTechnique = m_TechniqueIndex);

  Result := UpdateTechniqueSpecificVariables;
  // if FAILED(Result) then Exit;
end;


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HRESULT;
begin
  if Assigned(m_pFont) then m_pFont.OnLostDevice;
  if Assigned(m_pEffect) then m_pEffect.OnLostDevice;

  SAFE_RELEASE(m_pFullScreenTextureSurf);
  SAFE_RELEASE(m_pFullScreenTexture);
  SAFE_RELEASE(m_pRenderToSurface);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HRESULT;
begin
  SAFE_RELEASE(m_pFont);
  SAFE_RELEASE(m_pEffect);

  SAFE_RELEASE(m_pFullScreenTextureSurf);
  SAFE_RELEASE(m_pFullScreenTexture);
  SAFE_RELEASE(m_pRenderToSurface);

  SAFE_RELEASE(m_pScene1Mesh);
  SAFE_RELEASE(m_pScene1MeshTexture);
  SAFE_RELEASE(m_pScene2Mesh);
  SAFE_RELEASE(m_pScene2MeshTexture);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HRESULT;
begin
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HRESULT;
begin
  // Simple static scene, so just update the camera
  m_Camera.FrameMove(m_fElapsedTime);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------

function CMyD3DApplication.Render: HRESULT;
const
   mScene2WorldPos: array[0..2] of TD3DXVector3 = (
     (x:-0.5; y:-0.5; z:-0.5),
     (x: 1.0; y: 1.0; z: 2.0),
     (x: 3.0; y: 3.0; z: 5.0)
   );
var
  iPass, cPasses: LongWord;
  matWorld, matView, matProj, matViewProj: TD3DXMatrixA16;
  fLerp: Single;

  pSceneMesh: ID3DXMesh;
  nNumObjectsInScene: Integer;

  iObject: Integer;
  matRot, matPos: TD3DXMatrixA16;
  matWorldViewProj, matWorldView: TD3DXMatrixA16;
begin
  // First render the world on the rendertarget m_pFullScreenTexture.
  if SUCCEEDED(m_pRenderToSurface.BeginScene(m_pFullScreenTextureSurf, @m_ViewportOffscreen)) then
  begin
    m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, m_dwBackgroundColor, 1.0, 0);

    // Get the view & projection matrix from camera
    matView := m_Camera.GetViewMatrix^;
    matProj := m_Camera.GetProjMatrix^;
    D3DXMatrixMultiply(matViewProj, matView, matProj);

    // Update focal plane
    fLerp := (m_fTime - m_fChangeTime) / 2.0;
    fLerp := min(fLerp, 1.0);
    D3DXVec4Lerp(m_vFocalPlaneCur, m_vFocalPlaneSrc, m_vFocalPlaneDest, fLerp);
    m_pEffect.SetVector(m_hFocalPlane, m_vFocalPlaneCur);

    // Set world render technique
    m_pEffect.SetTechnique(m_hTechWorldWithBlurFactor);

    // Set the mesh texture
    if (m_nCurrentScene = 1) then
    begin
      m_pEffect.SetTexture(m_hMeshTexture, m_pScene1MeshTexture);
      pSceneMesh := m_pScene1Mesh;
      nNumObjectsInScene := 25;
    end else
    begin
      m_pEffect.SetTexture(m_hMeshTexture, m_pScene2MeshTexture);
      pSceneMesh := m_pScene2Mesh;
      nNumObjectsInScene := 3;
    end;

    for iObject:= 0 to nNumObjectsInScene - 1 do
    begin
      // setup the world matrix for the current world
      if (m_nCurrentScene = 1) then
      begin
        D3DXMatrixTranslation(matWorld, -(iObject mod 5)*1.0, 0.0, (iObject div 5)*3.0);
      end else
      begin
        D3DXMatrixRotationY(matRot, m_fTime * 0.66666);
        D3DXMatrixTranslation(matPos, mScene2WorldPos[iObject].x, mScene2WorldPos[iObject].y, mScene2WorldPos[iObject].z);
        D3DXMatrixMultiply(matWorld, matRot, matPos);
      end;

      // Update effect vars
      D3DXMatrixMultiply(matWorldViewProj, matWorld, matViewProj);
      D3DXMatrixMultiply(matWorldView, matWorld, matView);
      m_pEffect.SetMatrix(m_hWorld, matWorld);
      m_pEffect.SetMatrix(m_hWorldView, matWorldView);
      m_pEffect.SetMatrix(m_hWorldViewProjection, matWorldViewProj);

      // Draw the mesh on the rendertarget
      m_pEffect._Begin(@cPasses, 0);
      for iPass := 0 to cPasses - 1 do
      begin
        m_pEffect.Pass(iPass);
        pSceneMesh.DrawSubset(0);
      end;
      m_pEffect._End;
    end;

    m_pRenderToSurface.EndScene(0);
  end;

  // Clear the backbuffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET, $00000000, 1.0, 0);

  // Begin the scene, rendering to the backbuffer
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    m_pd3dDevice.SetViewport(m_ViewportFB);

    // Set the post process technique
    if (m_bShowBlurFactor) then
      m_pEffect.SetTechnique(m_hTechShowBlurFactor)
    else if (m_bShowUnblurred) then
      m_pEffect.SetTechnique(m_hTechShowUnmodified)
    else
      m_pEffect.SetTechnique(m_hTech[m_TechniqueIndex]);

    // Render the fullscreen quad on to the backbuffer
    m_pEffect._Begin(@cPasses, 0);
    for iPass := 0 to cPasses - 1 do
    begin
      m_pEffect.Pass(iPass);
      m_pd3dDevice.SetFVF(TVertex_FVF);
      m_pd3dDevice.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, m_Vertex, SizeOf(TVertex));
    end;
    m_pEffect._End;

    // Render the text
    RenderText;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RenderText()
// Desc: Draw the help & statistics for running sample
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.RenderText;
var
  rc: TRect;
  str: array[0..255] of Char;
  iLineY: Integer;
  tszTechName: PChar;
{$IFDEF UNICODE}
  tszBuf: array[0..MAX_PATH-1] of Char;
{$ENDIF}
begin
  if (nil = m_pFont) then Exit;

{$IFDEF UNICODE}
  For UNICODE you should use "m_pFont.DrawTextW" routine!!!
{$ENDIF}

  // Output statistics
  iLineY := 0;
  SetRect(rc, 2, iLineY, 0, 0); Inc(iLineY, 15);
  m_pFont.DrawTextA(nil, m_strFrameStats, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 0.0, 1.0)));
  SetRect(rc, 2, iLineY, 0, 0); Inc(iLineY, 15);
  m_pFont.DrawTextA( nil, m_strDeviceStats, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 0.0, 1.0)));

  if (m_bShowBlurFactor) then
    StrCopy(str, 'Technique: ShowBlurFactor')
  else if (m_bShowUnblurred) then
    StrCopy(str, 'Technique: ShowUnmodified')
  else
  begin
{$IFDEF UNICODE}
    tszTechName := tszBuf;
    MultiByteToWideChar(CP_ACP, 0, m_TechniqueNames[m_TechniqueIndex], -1, tszTechName, MAX_PATH);
    tszBuf[MAX_PATH - 1] :=  #0;
{$ELSE}
    tszTechName := PChar(m_TechniqueNames[m_TechniqueIndex]);
{$ENDIF}
    StrLFmt(str, 256, 'Technique: %s', [tszTechName]);
  end;
  SetRect(rc, 2, iLineY, 0, 0); Inc(iLineY, 15);
  m_pFont.DrawTextA(nil, str, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));

  StrLFmt(str, 256, 'Blur Factor: %0.1f  Focal Plane: (%0.1f,%0.1f,%0.1f,%0.1f)', [m_fBlurConst, m_vFocalPlaneCur.x, m_vFocalPlaneCur.y, m_vFocalPlaneCur.z, m_vFocalPlaneCur.w]);
  SetRect(rc, 2, iLineY, 0, 0); Inc(iLineY, 15);
  m_pFont.DrawTextA(nil, str, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));

  if (m_bShowHelp) then
  begin
    iLineY := m_d3dsdBackBuffer.Height-15*6;
    SetRect(rc, 2, iLineY, 0, 0); Inc(iLineY, 15);
    m_pFont.DrawTextA(nil, 'Controls:', -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 0.75, 0.0, 1.0)));
    SetRect(rc, 20, iLineY, 0, 0);
    m_pFont.DrawTextA(nil, 'Look: Left drag mouse'#10 +
                           'Move: A,W,S,D or Arrow Keys'#10 +
                           'Strafe up/down: Q,E or PgUp,PgDn'#10 +
                           'Change Scene: P'#10 +
                           'Change Technique: Plus,Minus'#10,
                           -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 0.75, 0.0, 1.0)));

    SetRect(rc, 250, iLineY, 0, 0);
    m_pFont.DrawTextA(nil, 'Show Blur Factor (Alpha): I'#10 +
                           'Show Unblurred: U'#10 +
                           'Change Blur Factor: Y,H,N'#10 +
                           'Change Focal Plane: 1,2,3'#10 +
                           'Change Blur Factor: Insert,Delete,R'#10,
                           -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 0.75, 0.0, 1.0)));

    SetRect(rc, 500, iLineY, 0, 0);
    m_pFont.DrawTextA(nil, 'Hide help: F1'#10 +
                           'Change Device: F2'#10 +
                           'View readme: F9'#10 +
                           'Quit: ESC'#10,
                           -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 0.75, 0.0, 1.0)));
  end else
  begin
    SetRect(rc, 2, iLineY, 0, 0); // Inc(iLineY, 15); //BAA: never used anyway
    m_pFont.DrawTextA(nil, 'Press F1 for help',
                      -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
  end;
end;





//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9;
  dwBehavior: DWORD; adapterFormat, backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  // Debugging vertex shaders requires either REF or software vertex processing
  // and debugging pixel shaders requires REF.
  {$IFDEF DEBUG_VS}
    if (pCaps.DeviceType <> D3DDEVTYPE_REF) and
       ((dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING) = 0) then Exit;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
    if (pCaps.DeviceType <> D3DDEVTYPE_REF) then Exit;
  {$ENDIF}

  if (dwBehavior and D3DCREATE_PUREDEVICE <> 0) then Exit;

  if (pCaps.PixelShaderVersion < D3DPS_VERSION(1,1)) then Exit;

  // If device doesn't support 1.1 vertex shaders in HW, switch to SWVP.
  if (pCaps.VertexShaderVersion < D3DVS_VERSION(1,1)) then
  begin
    if ((dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING) = 0) then Exit;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: LongWord;
  wParam: WPARAM; lParam: LPARAM): LRESULT;
var
  vFocalPlane: TD3DXVector4;
  OriginalTechnique: DWORD;
  hTech: TD3DXHandle;
  vecEye, vecAt: TD3DXVector3;
const
  IsChecked: array[False..True] of Integer = (MF_UNCHECKED, MF_CHECKED);
begin
  // Pass messages to camera class
  m_Camera.HandleMessages(hWnd, uMsg, wParam, lParam);

  case uMsg of
    WM_COMMAND:
    begin
      m_pEffect.GetVector(m_hFocalPlane, vFocalPlane);

      case LOWORD(wParam) of
        IDM_OPTIONS_SHOWBLURFACTOR:
        begin
          m_bShowBlurFactor := not m_bShowBlurFactor;
          if (m_bShowBlurFactor) then  m_bShowUnblurred := False;
          CheckMenuItem(GetMenu(hWnd), IDM_OPTIONS_SHOWBLURFACTOR, IsChecked[m_bShowBlurFactor]);
          CheckMenuItem(GetMenu(hWnd), IDM_OPTIONS_SHOWUNBLURRED, IsChecked[m_bShowUnblurred]);
        end;

        IDM_OPTIONS_SHOWUNBLURRED:
        begin
          m_bShowUnblurred := not m_bShowUnblurred;
          if (m_bShowUnblurred) then m_bShowBlurFactor := False;
          CheckMenuItem(GetMenu(hWnd), IDM_OPTIONS_SHOWBLURFACTOR, IsChecked[m_bShowBlurFactor]);
          CheckMenuItem(GetMenu(hWnd), IDM_OPTIONS_SHOWUNBLURRED, IsChecked[m_bShowUnblurred]);
        end;

        IDM_OPTIONS_NEXTTECHNIQUE:
        begin
          OriginalTechnique := m_TechniqueIndex;
          repeat
            Inc(m_TechniqueIndex);
            if (m_TechniqueIndex = m_TechniqueCount) then m_TechniqueIndex := 0;

            hTech := m_pEffect.GetTechniqueByName(m_TechniqueNames[m_TechniqueIndex]);
            if SUCCEEDED(m_pEffect.ValidateTechnique(hTech)) then Break;
          until (OriginalTechnique = m_TechniqueIndex);

          UpdateTechniqueSpecificVariables;
        end;

        IDM_OPTIONS_PREVTECHNIQUE:
        begin
          OriginalTechnique := m_TechniqueIndex;
          repeat
            if (m_TechniqueIndex = 0) then m_TechniqueIndex := m_TechniqueCount;
            Dec(m_TechniqueIndex);

//todo: Fill BUG REPORT            
            hTech := m_pEffect.GetTechniqueByName(m_TechniqueNames[m_TechniqueIndex]);
            if SUCCEEDED(m_pEffect.ValidateTechnique(hTech)) then Break;
//              if (SUCCEEDED(m_pEffect->ValidateTechnique(m_TechniqueNames[m_TechniqueIndex])))
          until (OriginalTechnique = m_TechniqueIndex);

          UpdateTechniqueSpecificVariables;
        end;

        IDM_OPTIONS_FOCALPLANE1:
        begin
          m_fChangeTime := m_fTime;
          m_vFocalPlaneDest := D3DXVector4(0.0, 0.0, 1.0, -2.5);
          m_vFocalPlaneSrc := vFocalPlane;
        end;

        IDM_OPTIONS_FOCALPLANE2:
        begin
          m_fChangeTime := m_fTime;
          m_vFocalPlaneDest := D3DXVector4(0.0, 0.0, 1.0, -5.0);
          m_vFocalPlaneSrc := vFocalPlane;
        end;

        IDM_OPTIONS_FOCALPLANE3:
        begin
          m_fChangeTime := m_fTime;
          m_vFocalPlaneDest := D3DXVector4(0.0, 0.0, 1.0, -8.0);
          m_vFocalPlaneSrc := vFocalPlane;
        end;

        IDM_OPTIONS_INCREASEBLUR:
        begin
          m_fBlurConst := m_fBlurConst + 0.25;
          UpdateTechniqueSpecificVariables;
        end;

        IDM_OPTIONS_DECREASEBLUR:
        begin
          m_fBlurConst := m_fBlurConst - 0.25;
          if (m_fBlurConst < 0) then m_fBlurConst := 0; // bottoms at 0
          UpdateTechniqueSpecificVariables;
        end;

        IDM_OPTIONS_RESETBLUR:
        begin
          m_fBlurConst := 4.0;
          UpdateTechniqueSpecificVariables;
        end;

        IDM_OPTIONS_CHANGESCENE:
        begin
          m_nCurrentScene := m_nCurrentScene mod 2;
          Inc(m_nCurrentScene);

          case m_nCurrentScene of
            1:
            begin
              vecEye := D3DXVector3(0.75, 0.8, -2.3);
              vecAt  := D3DXVector3(0.2, 0.75, -1.5);
              m_Camera.SetViewParams(vecEye, vecAt);
            end;

            2:
            begin
              vecEye := D3DXVector3(0.0, 0.0, -3.0);
              vecAt  := D3DXVector3(0.0, 0.0, 0.0);
              m_Camera.SetViewParams(vecEye, vecAt);
            end;
          end;
        end;

        IDM_TOGGLEHELP: m_bShowHelp := not m_bShowHelp;
      end;
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.

